﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media.Animation;
using Microsoft.Phone.Controls;

namespace MiddlemarchReader
{
    public partial class BookViewer : UserControl
    {
        Border[] pageHosts = new Border[3];
        int pageHostBaseIndex = 0;          // current page
        double dragFactor;

        int nextPageChapterIndex;
        int nextPagePageIndex;
        int prevPageChapterIndex;
        int prevPagePageIndex;

        bool animationInProgress, gestureInProgress, gestureHadFlick, canGoNextPage, canGoPrevPage;

        public static readonly DependencyProperty PageProviderProperty =
            DependencyProperty.Register("PageProvider",
                typeof(IPageProvider),
                typeof(BookViewer),
                new PropertyMetadata(OnPageProviderChanged));

        public static readonly DependencyProperty PageTransitionProperty =
            DependencyProperty.Register("PageTransition",
                typeof(PageTransition),
                typeof(BookViewer),
                new PropertyMetadata(OnPageTransitionChanged));

        public static readonly DependencyProperty ChapterIndexProperty =
            DependencyProperty.Register("ChapterIndex",
                typeof(int),
                typeof(BookViewer),
                new PropertyMetadata(0, OnPageChanged));

        public static readonly DependencyProperty PageIndexProperty =
            DependencyProperty.Register("PageIndex",
                typeof(int),
                typeof(BookViewer),
                new PropertyMetadata(0, OnPageChanged));

        public event EventHandler PageChanged; 

        public BookViewer()
        {
            InitializeComponent();

            pageHosts[0] = pageHost0;
            pageHosts[1] = pageHost1;
            pageHosts[2] = pageHost2;
        }

        public IPageProvider PageProvider
        {
            set { SetValue(PageProviderProperty, value); }
            get { return (IPageProvider)GetValue(PageProviderProperty); }
        }

        public PageTransition PageTransition
        {
            set { SetValue(PageTransitionProperty, value); }
            get { return (PageTransition)GetValue(PageTransitionProperty); }
        }

        public int ChapterIndex
        {
            protected set { SetValue(ChapterIndexProperty, value); }
            get { return (int)GetValue(ChapterIndexProperty); }
        }

        public int PageIndex
        {
            protected set { SetValue(PageIndexProperty, value); }
            get { return (int)GetValue(PageIndexProperty); }
        }

        public bool JumpToPage(int chapterIndex, int pageIndex)
        {
            // Save chapter and page indices in public properties
            this.ChapterIndex = chapterIndex;
            this.PageIndex = pageIndex;

            if (this.PageProvider != null)
            {
                this.PageProvider.SetPageSize(new Size(pageHost0.ActualWidth, pageHost0.ActualHeight));
                this.PageProvider.SetFont(this.FontFamily, this.FontSize);
                FrameworkElement element = this.PageProvider.GetPage(chapterIndex, pageIndex);

                if (element != null)
                {
                    // Set pageHosts for this page and the next page
                    pageHosts[pageHostBaseIndex].Child = element;
                    SetUpNextPage();
                    SetUpPreviousPage();

                    // Set the FractionlBaseIndex for the PageTransition
                    this.PageTransition.FractionalBaseIndex = pageHostBaseIndex;
                    return true;
                }
            }
            return false;
        }

        void OnGestureListenerGestureBegin(object sender, GestureEventArgs args)
        {
            gestureInProgress = !animationInProgress;

            if (gestureInProgress)
            {
                gestureHadFlick = false;
                canGoNextPage = pageHosts[(pageHostBaseIndex + 1) % 3].Child != null;
                canGoPrevPage = pageHosts[(pageHostBaseIndex + 2) % 3].Child != null;
            }
        }

        void OnGestureListenerGestureCompleted(object sender, GestureEventArgs args)
        {
            gestureInProgress = false;
        }

        void OnGestureListenerTap(object sender, GestureEventArgs args)
        {
            if (gestureInProgress && canGoNextPage)
            {
                AnimateToNextPage();
            }
        }

        void OnGestureListenerFlick(object sender, FlickGestureEventArgs args)
        {
            if (gestureInProgress)
            {
                if (args.Angle < 45 || args.Angle > 315)
                {
                    if (!canGoPrevPage)
                    {
                        if (this.ChapterIndex != 0 && this.PageIndex == 0)
                        {
                            pageHosts[(pageHostBaseIndex + 2) % 3].Child = PageProvider.GetLastPage(prevPageChapterIndex, out prevPagePageIndex);
                            canGoPrevPage = pageHosts[(pageHostBaseIndex + 2) % 3].Child != null;
                        }
                    }
                    if (canGoPrevPage)
                    {
                        gestureHadFlick = true;
                        AnimateToPreviousPage();
                    }
                }

                else if (args.Angle > 135 && args.Angle < 225 && canGoNextPage)
                {
                    gestureHadFlick = true;
                    AnimateToNextPage();
                }
            }
        }

        void OnGestureListenerDragStarted(object sender, DragStartedGestureEventArgs args)
        {
            dragFactor = 0;
        }

        void OnGestureListenerDragDelta(object sender, DragDeltaGestureEventArgs args)
        {
            if (gestureInProgress)
            {
                double newDragFactor = dragFactor - args.HorizontalChange / pageHosts[pageHostBaseIndex].ActualWidth;

                if (newDragFactor < 0 && !canGoPrevPage && this.ChapterIndex != 0 && this.PageIndex == 0)
                {
                    pageHosts[(pageHostBaseIndex + 2) % 3].Child = PageProvider.GetLastPage(prevPageChapterIndex, out prevPagePageIndex);
                    canGoPrevPage = pageHosts[(pageHostBaseIndex + 2) % 3].Child != null;
                }

                double maximum = canGoNextPage ? 1 : 0;
                double minimum = canGoPrevPage ? -1 : 0;
                dragFactor = Math.Max(minimum, Math.Min(maximum, newDragFactor));
                this.PageTransition.FractionalBaseIndex = pageHostBaseIndex + dragFactor;
            }
        }

        void OnGestureListenerDragCompleted(object sender, DragCompletedGestureEventArgs args)
        {
            if (gestureInProgress && !gestureHadFlick)
            {
                if (dragFactor > 0.5)
                    AnimateToNextPage();

                else if (dragFactor < -0.5)
                    AnimateToPreviousPage();

                else AnimateToSamePage();
            }
        }

        void SetUpNextPage()
        {
            nextPageChapterIndex = this.ChapterIndex;
            nextPagePageIndex = this.PageIndex + 1;
            FrameworkElement element = this.PageProvider.GetPage(nextPageChapterIndex, nextPagePageIndex);

            if (element == null)
            {
                nextPageChapterIndex += 1;
                nextPagePageIndex = 0;
                element = this.PageProvider.GetPage(nextPageChapterIndex, nextPagePageIndex);
            }

            pageHosts[(pageHostBaseIndex + 1) % 3].Child = element;
        }

        void SetUpPreviousPage()
        {
            prevPageChapterIndex = this.ChapterIndex;
            prevPagePageIndex = this.PageIndex - 1;
            FrameworkElement element = null;

            if (prevPagePageIndex != -1)
            {
                element = this.PageProvider.GetPage(prevPageChapterIndex, prevPagePageIndex);
            }
            else if (prevPageChapterIndex != 0)
            {
                prevPageChapterIndex -= 1;
                element = this.PageProvider.QueryLastPage(prevPageChapterIndex, out prevPagePageIndex);
            }
            pageHosts[(pageHostBaseIndex + 2) % 3].Child = element;
        }

        void AnimateToNextPage()
        {
            DoubleAnimation anima = new DoubleAnimation
            {
                To = pageHostBaseIndex + 1,
                Duration = CalculateFractionalDuration(pageHostBaseIndex + 1)
            };

            Storyboard.SetTarget(anima, this.PageTransition);
            Storyboard.SetTargetProperty(anima, new PropertyPath(PageTransition.FractionalBaseIndexProperty));

            Storyboard storyboard = new Storyboard();
            storyboard.Children.Add(anima);

            storyboard.Completed += (object sender, EventArgs args) =>
                {
                    pageHostBaseIndex = (pageHostBaseIndex + 1) % 3;
                    this.PageTransition.FractionalBaseIndex = pageHostBaseIndex;

                    // Shift the indices from current to previous
                    prevPageChapterIndex = this.ChapterIndex;
                    prevPagePageIndex = this.PageIndex;

                    // Shift the indices from next to current
                    this.ChapterIndex = nextPageChapterIndex;
                    this.PageIndex = nextPagePageIndex;

                    // Set up the new next page
                    SetUpNextPage();
                    animationInProgress = false;
                };

            animationInProgress = true;
            storyboard.Begin();
        }

        void AnimateToPreviousPage()
        {
            DoubleAnimation anima = new DoubleAnimation
            {
                To = pageHostBaseIndex - 1,
                Duration = CalculateFractionalDuration(pageHostBaseIndex - 1)
            };

            Storyboard.SetTarget(anima, this.PageTransition);
            Storyboard.SetTargetProperty(anima, new PropertyPath(PageTransition.FractionalBaseIndexProperty));

            Storyboard storyboard = new Storyboard();
            storyboard.Children.Add(anima);

            storyboard.Completed += (object sender, EventArgs args) =>
            {
                pageHostBaseIndex = (pageHostBaseIndex + 2) % 3;
                this.PageTransition.FractionalBaseIndex = pageHostBaseIndex;

                // Transfer the indices from current to next
                nextPageChapterIndex = this.ChapterIndex;
                nextPagePageIndex = this.PageIndex;

                // Transfer the indices from previous to current
                this.ChapterIndex = prevPageChapterIndex;
                this.PageIndex = prevPagePageIndex;

                // Set up the new previous page
                SetUpPreviousPage();
                animationInProgress = false;
            };

            animationInProgress = true;
            storyboard.Begin();
        }

        void AnimateToSamePage()
        {
            DoubleAnimation anima = new DoubleAnimation
            {
                To = pageHostBaseIndex,
                Duration = CalculateFractionalDuration(pageHostBaseIndex)
            };

            Storyboard.SetTarget(anima, this.PageTransition);
            Storyboard.SetTargetProperty(anima, new PropertyPath(PageTransition.FractionalBaseIndexProperty));

            Storyboard storyboard = new Storyboard();
            storyboard.Children.Add(anima);

            storyboard.Completed += (object sender, EventArgs args) =>
            {
                animationInProgress = false;
            };

            animationInProgress = true;
            storyboard.Begin();
        }

        Duration CalculateFractionalDuration(double fractionalBaseIndexDestination)
        {
            double fraction = Math.Abs(fractionalBaseIndexDestination - 
                                       this.PageTransition.FractionalBaseIndex);
            double msec = fraction * this.PageTransition.AnimationDuration;
            TimeSpan timeSpan = TimeSpan.FromMilliseconds(msec);
            return new Duration(timeSpan);
        }

        static void OnPageProviderChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
        }

        static void OnPageTransitionChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as BookViewer).OnPageTransitionChanged((PageTransition)args.OldValue, (PageTransition)args.NewValue);
        }

        void OnPageTransitionChanged(PageTransition oldPageTransition, PageTransition newPageTransition)
        {
            double fractionalBaseIndex = -1;

            if (oldPageTransition != null)
            {
                fractionalBaseIndex = oldPageTransition.FractionalBaseIndex;
                oldPageTransition.Detach();
            }
            if (newPageTransition != null)
            {
                newPageTransition.Attach(LayoutRoot, pageContainer0, pageContainer1, pageContainer2);
                newPageTransition.FractionalBaseIndex = fractionalBaseIndex;
            }
        }

        static void OnPageChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            BookViewer bookViewer = obj as BookViewer;

            if (bookViewer.PageChanged != null)
                bookViewer.PageChanged(bookViewer, EventArgs.Empty);
        }
    }
}
