﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Shapes;

namespace MiddlemarchReader
{
    public class CurlTransition : PageTransition
    {
        const bool curlFromTop = true;              // calculations really assume that bottom right corner curls
        const double R1 = 1.0 / 3;                  // portion of circle radius above bottom (or below top) of page as fraction of page height
        const double R = (1 + R1 * R1) / (2 * R1);  // radius of arc governing the passage of the lower right (or upper right) corner

        bool isAttached;
        Panel containerPanel;
        FrameworkElement[] pageContainers = new FrameworkElement[3];

        // Clipping geometry
        PathGeometry pathGeometry = new PathGeometry();
        PathFigure pathFigure = new PathFigure { IsClosed = true };
        PolyLineSegment polylineSegment = new PolyLineSegment();
        int clipGeometryOwner = -1;

        // Back of page polygon
        Polygon polygon = new Polygon
        {
            Stroke = new SolidColorBrush(Colors.Black),
            Fill = new SolidColorBrush(Color.FromArgb(255, 224, 224, 224)),
            Visibility = Visibility.Collapsed
        };

        public override void Attach(Panel containerPanel, 
                                    FrameworkElement pageContainer0, 
                                    FrameworkElement pageContainer1, 
                                    FrameworkElement pageContainer2)
        {
            this.containerPanel = containerPanel;
            pageContainers[0] = pageContainer0;
            pageContainers[1] = pageContainer1;
            pageContainers[2] = pageContainer2;

            pathFigure.Segments.Add(polylineSegment);
            pathGeometry.Figures.Add(pathFigure);

            // Keep Polygon on top of all page containers
            Canvas.SetZIndex(polygon, 4);
            containerPanel.Children.Add(polygon);

            isAttached = true;
        }

        public override void Detach()
        {
            if (isAttached)
            {
                for (int i = 0; i < 3; i++)
                    pageContainers[i].Clip = null;

                containerPanel.Children.Remove(polygon);
                isAttached = false;
            }
        }

        protected override void ShowPageTransition(int baseIndex, double fraction)
        {
            int nextIndex = (baseIndex + 1) % 3;
            int prevIndex = (baseIndex + 2) % 3;
            int curlIndex = 0;
            double angle = Math.Asin(1 / R);    // Initial angle traveling along arc

            if (!isAttached)
                return;

            if (fraction >= 0)
            {
                Canvas.SetZIndex(pageContainers[prevIndex], 1);
                Canvas.SetZIndex(pageContainers[nextIndex], 2);
                Canvas.SetZIndex(pageContainers[baseIndex], 3);
                curlIndex = baseIndex;
                angle *= 1 - 2 * fraction;
            }
            else  // fraction < 0
            {
                Canvas.SetZIndex(pageContainers[nextIndex], 1);
                Canvas.SetZIndex(pageContainers[baseIndex], 2);
                Canvas.SetZIndex(pageContainers[prevIndex], 3);
                curlIndex = prevIndex;
                angle *= -1 - 2 * fraction;
            }

            if (fraction == 0)
            {
                if (clipGeometryOwner != -1)
                {
                    pageContainers[clipGeometryOwner].Clip = null;
                    clipGeometryOwner = -1;
                }
                polygon.Visibility = Visibility.Collapsed;
            }
            else
            {
                double width = pageContainers[curlIndex].ActualWidth;
                double height = pageContainers[curlIndex].ActualHeight;

                // Calculate coordinates of corner traveling along arc
                double xArc = width * R * Math.Sin(angle);
                double yArc = height * (1 + (R - R1) - R * Math.Cos(angle));

                // Apply Pythagorean theorems to get bottom and righ intersects
                double xBottom = (width * width - xArc * xArc - Math.Pow(height - yArc, 2)) / (2 * (width - xArc));
                double yRight = (height * height - yArc * yArc - Math.Pow(width - xArc, 2)) / (2 * (height - yArc));

                // These two factors govern alter Y coordinates so the page curls from the top
                double minuend = curlFromTop ? height : 0;
                double multiplier = curlFromTop ? -1 : 1;

                // The clipping path
                polylineSegment.Points.Clear();
                pathFigure.StartPoint = new Point(0, minuend);

                // The back of the page polygon
                polygon.Points.Clear();

                if (!Double.IsInfinity(yRight))
                {
                    if (yRight > 0)
                    {
                        polylineSegment.Points.Add(new Point(width, minuend));
                        polylineSegment.Points.Add(new Point(width, minuend + multiplier * yRight));

                        polygon.Points.Add(new Point(width, minuend + multiplier * yRight));
                    }
                    else // the corner of the page has gone above the top (or below the bottom)
                    {
                        // Use similar triangles to calculate intersect on top of page
                        double xTop = ((width - xBottom) * yRight) / (height - yRight) + width;
                        polylineSegment.Points.Add(new Point(xTop, minuend));

                        // Use a vector to calculate corner of page above the top
                        Point vector = new Point(width - xArc, yRight - yArc);
                        double length = Math.Sqrt(vector.X * vector.X + vector.Y * vector.Y);
                        double xCorner = xArc + height * vector.X / length;
                        double yCorner = yArc + height * vector.Y / length;

                        polygon.Points.Add(new Point(xTop, minuend));
                        polygon.Points.Add(new Point(xCorner, minuend + multiplier * yCorner));
                    }

                    polylineSegment.Points.Add(new Point(xBottom, minuend + multiplier * height));
                    polylineSegment.Points.Add(new Point(0, minuend + multiplier * height));

                    if (clipGeometryOwner != -1 && clipGeometryOwner != curlIndex)
                        pageContainers[clipGeometryOwner].Clip = null;

                    pageContainers[curlIndex].Clip = pathGeometry;
                    clipGeometryOwner = curlIndex;

                    polygon.Points.Add(new Point(xArc, minuend + multiplier * yArc));
                    polygon.Points.Add(new Point(xBottom, minuend + multiplier * height));
                    polygon.Visibility = Visibility.Visible;
                }
            }
        }
    }
}
