﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.IO;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;

namespace MiddlemarchReader
{
    public class PageProvider : IPageProvider
    {
        static readonly Brush blackBrush = new SolidColorBrush(Colors.Black);

        List<string> paragraphs = new List<string>();
        Size pageSize;
        FontFamily fontFamily;
        double fontSize;

        public BookInfo Book { set; get; }

        protected void PreprocessBook(Stream stream)
        {
            // Occurs every time program is run
            GenerateParagraphs(stream);

            // Occurs only first time program is run
            if (this.Book.Chapters.Count == 0)
                GenerateChapters();
        }

        void GenerateParagraphs(Stream stream)
        {
            using (StreamReader reader = new StreamReader(stream))
            {
                string line;
                StringBuilder paragraph = new StringBuilder();

                while (null != (line = reader.ReadLine()))
                {
                    // Check for a possible paragraph termination
                    if (line.Length == 0 || line[0] == ' ')
                    {
                        if (paragraph.Length > 0)
                        {
                            paragraphs.Add(paragraph.ToString());
                            paragraph.Remove(0, paragraph.Length);
                        }

                        // Add the line anyway
                        paragraphs.Add(line);
                    }
                    else
                    {
                        // Insert a space between appended lines assembled into paragraph
                        if (paragraph.Length > 0)
                            paragraph.Append(" ");

                        paragraph.Append(line);
                    }
                }

                // Add the last paragraph if there's one in progress
                if (paragraph.Length > 0)
                    paragraphs.Add(paragraph.ToString());
            }
        }

        void GenerateChapters()
        {
            // Set so the beginning becomes the first chapter
            int consecutiveBlankLineCount = 10;
            int accumulatedCharacterCount = 0;

            for (int i = 0; i < paragraphs.Count; i++)
            {
                if (paragraphs[i].Length == 0)
                {
                    consecutiveBlankLineCount += 1;
                }
                else
                {
                    if (consecutiveBlankLineCount > 2)
                    {
                        // Create a new chapter
                        PageInfo pageInfo = new PageInfo 
                        { 
                            ParagraphIndex = i, 
                            CharacterIndex = 0, 
                            AccumulatedCharacterCount = accumulatedCharacterCount 
                        };

                        ChapterInfo chapterInfo = new ChapterInfo 
                        { 
                            Title = paragraphs[i] 
                        };
                        chapterInfo.Pages.Add(pageInfo);
                        this.Book.Chapters.Add(chapterInfo);
                    }
                    consecutiveBlankLineCount = 0;
                    accumulatedCharacterCount += paragraphs[i].Length;
                }
            }

            // Flag the last chapter
            this.Book.Chapters[this.Book.Chapters.Count - 1].IsLastChapter = true;
            this.Book.CharacterCount = accumulatedCharacterCount;
        }

        public void SetPageSize(Size size) 
        {
            if (pageSize != size)
            {
                pageSize = size;

                // TODO: If the page size can really change,
                //      pagination information needs to be invalidated
            }
        }

        public void SetFont(FontFamily fontFamily, double fontSize)
        {
            this.fontFamily = fontFamily;
            this.fontSize = fontSize;

            // TODO: If the font and font size can really change,
            //      pagination information needs to be invalidated
        }


        public FrameworkElement QueryLastPage(int chapterIndex, out int pageIndex) 
        {
            return GetLastPage(chapterIndex, out pageIndex, false);
        }

        public FrameworkElement GetLastPage(int chapterIndex, out int pageIndex)
        {
            return GetLastPage(chapterIndex, out pageIndex, true);
        }

        FrameworkElement GetLastPage(int chapterIndex, out int pageIndex, bool paginateIfNecessary)
        {
            if (pageSize.IsEmpty)
            {
                pageIndex = -1;
                return null;
            }

            if (chapterIndex < 0 || chapterIndex > this.Book.Chapters.Count - 1)
            {
                pageIndex = -1;
                return null;
            }

            ChapterInfo chapter = this.Book.Chapters[chapterIndex];

            if (chapter.Pages[chapter.Pages.Count - 1].IsLastPage)
            {
                pageIndex = chapter.Pages.Count - 1;
                PageInfo nextPageInfo = chapter.IsLastChapter ? null : this.Book.Chapters[chapterIndex + 1].Pages[0];
                return GetPaginatedPage(chapter.Pages[chapter.Pages.Count - 1], nextPageInfo);
            }

            if (!paginateIfNecessary)
            {
                pageIndex = -1;
                return null;
            }

            FrameworkElement element = null;

            while (!chapter.Pages[chapter.Pages.Count - 1].IsLastPage)
                element = GetPage(chapterIndex, chapter.Pages.Count - 1);

            pageIndex = chapter.Pages.Count - 1;
            return element;
        }

        public FrameworkElement GetPage(int chapterIndex, int pageIndex)
        {
            if (pageSize.IsEmpty)
                return null;

            if (pageIndex < 0)
                return null;

            if (chapterIndex < 0 || chapterIndex > this.Book.Chapters.Count - 1)
                return null;

            ChapterInfo chapter = this.Book.Chapters[chapterIndex];

            // Check if the last page of the paragraph has already been established, 
            //  and the requested page is greater than that last page
            if (chapter.Pages[chapter.Pages.Count - 1].IsLastPage && pageIndex > chapter.Pages.Count - 1)
                return null;

            // If the last page of a chapter is requested, and it's already available, get it
            if (chapter.Pages[pageIndex].IsLastPage)
            {
                // Last chapter in book and hence last page in book
                if (chapterIndex == this.Book.Chapters.Count - 1)
                    return GetPaginatedPage(chapter.Pages[pageIndex], null);

                // The last page of a chapter but not the last chapter
                return GetPaginatedPage(chapter.Pages[pageIndex], this.Book.Chapters[chapterIndex + 1].Pages[0]);
            }

            // Otherewise, if the page has already been paginated, get it
            if (pageIndex + 1 < chapter.Pages.Count)
            {
                return GetPaginatedPage(chapter.Pages[pageIndex], chapter.Pages[pageIndex + 1]);
            }

            if (pageIndex != chapter.Pages.Count - 1)
                throw new Exception(String.Format("GetPage can only paginate one page per call: pageIndex = {0} while chapter.Pages.Count = {1}", pageIndex, chapter.Pages.Count));

            StackPanel stackPanel = new StackPanel();
            PageInfo pageInfo = chapter.Pages[pageIndex];
            int paragraphIndex = pageInfo.ParagraphIndex;
            int characterIndex = pageInfo.CharacterIndex;
            int accumulatedCharacterCount = pageInfo.AccumulatedCharacterCount;

            // Get the paragraph index of the next chapter, if it exists
            int nextChapterParagraphIndex = paragraphs.Count;

            if (chapterIndex + 1 < this.Book.Chapters.Count)
                nextChapterParagraphIndex = this.Book.Chapters[chapterIndex + 1].Pages[0].ParagraphIndex;

            while (paragraphIndex < paragraphs.Count && paragraphIndex < nextChapterParagraphIndex)
            {
                // Skip if a blank line is the first paragraph on a page
                if (stackPanel.Children.Count == 0 &&
                    characterIndex == 0 &&
                    paragraphs[paragraphIndex].Length == 0)
                {
                    paragraphIndex++;
                    continue;
                }

                TextBlock txtblk = CreateTextBlock(paragraphs[paragraphIndex].Substring(characterIndex));
                stackPanel.Children.Add(txtblk);
                stackPanel.Measure(new Size(pageSize.Width, Double.PositiveInfinity));

                // Check if the StackPanel fits in the available height
                if (stackPanel.DesiredSize.Height > pageSize.Height)
                {
                    // Strip words off the end until it fits
                    do
                    {
                        int index = txtblk.Text.LastIndexOf(' ');

                        if (index == -1)
                            index = 0;

                        txtblk.Text = txtblk.Text.Substring(0, index);
                        stackPanel.Measure(new Size(pageSize.Width, Double.PositiveInfinity));

                        if (index == 0)
                            break;
                    }
                    while (stackPanel.DesiredSize.Height > pageSize.Height);

                    characterIndex += txtblk.Text.Length;
                    accumulatedCharacterCount += txtblk.Text.Length;

                    // Skip over the space
                    if (txtblk.Text.Length > 0)
                        characterIndex++;

                    break;
                }
                accumulatedCharacterCount += txtblk.Text.Length;
                paragraphIndex++;
                characterIndex = 0;
            }

            // Skip over blank paragraphs. We don't want these at the top
            //   of the next page!
            while (paragraphIndex < nextChapterParagraphIndex)
            {
                if (paragraphs[paragraphIndex].Length == 0)
                    paragraphIndex++;
                else
                    break;
            }

            pageInfo.IsLastPage = paragraphIndex == nextChapterParagraphIndex;

            // Create a PageInfo for the next page only if it's not the
            //  last page in the chapter
            if (!pageInfo.IsLastPage)
            {
                PageInfo nextPageInfo = new PageInfo
                {
                    ParagraphIndex = paragraphIndex,
                    CharacterIndex = characterIndex,
                    AccumulatedCharacterCount = accumulatedCharacterCount
                };
                chapter.Pages.Add(nextPageInfo);
            }

            return stackPanel;
        }

        // Creates a page from one PageInfo to another.
        // NOTE: pageInfo2 is null if this is the last page of the book.
        FrameworkElement GetPaginatedPage(PageInfo pageInfo1, PageInfo pageInfo2)
        {
            Panel pnl = new StackPanel();

            // This is the case where a page has only part of a single paragraph
            if (pageInfo2 != null && pageInfo1.ParagraphIndex == pageInfo2.ParagraphIndex)
            {
                string text = paragraphs[pageInfo1.ParagraphIndex].Substring(pageInfo1.CharacterIndex, pageInfo2.CharacterIndex - pageInfo1.CharacterIndex);
                pnl.Children.Add(CreateTextBlock(text));
            }
            else
            {
                // Add first (possibly) partial paragraph
                string text = paragraphs[pageInfo1.ParagraphIndex].Substring(pageInfo1.CharacterIndex);
                pnl.Children.Add(CreateTextBlock(text));

                // Set paragraphIndex2 based on  pageInfo2 possibly being null
                int paragraphIndex2 = pageInfo2 != null ? pageInfo2.ParagraphIndex : paragraphs.Count;

                // Add full paragraphs
                for (int paragraphIndex = pageInfo1.ParagraphIndex + 1; paragraphIndex < paragraphIndex2; paragraphIndex++)
                {
                    text = paragraphs[paragraphIndex];
                    pnl.Children.Add(CreateTextBlock(text));
                }

                // Add partial paragraph at end
                if (pageInfo2 != null)
                {
                    text = paragraphs[pageInfo2.ParagraphIndex].Substring(0, pageInfo2.CharacterIndex);
                    pnl.Children.Add(CreateTextBlock(text));
                }
            }
            return pnl;
        }

        TextBlock CreateTextBlock(string text)
        {
            bool isBlankLine = String.IsNullOrEmpty(text);

            return new TextBlock
            {
                FontFamily = this.fontFamily,
                FontSize = this.fontSize,
                Text = isBlankLine ? " " : text,
                TextWrapping = TextWrapping.Wrap,
                Foreground = blackBrush,
                LineStackingStrategy = isBlankLine ? LineStackingStrategy.BlockLineHeight : LineStackingStrategy.MaxHeight,
                LineHeight = fontSize / (isBlankLine ? 2 : 1)
            };
        }
    }
}

